# -*- coding: utf-8 -*-

"""
Syntax of DDB table altering with SQL:
-------------------------------------
ALTER TABLE tbl_name (
    [col_name column_definition, ...]
    [alter_option [, alter_option] ...]
)
    [table_options]

alter_option: {
    CREATE INDEX [index_name] GLOBAL
        (col_name key_type, ...)
        [index_option] ...
    | UPDATE INDEX [index_name] GLOBAL
        [index_option] ...
    | DELETE INDEX [index_name] GLOBAL
    | CREATE REPLICA [region_name]
        [replica_option] ...
    | UPDATE REPLICA [region_name]
        [replica_option] ...
    | DELETE REPLICA [region_name]
}

key_type : {
    PARTITION KEY | SORT KEY | HASH | RANGE
}

index_option: {
    Projection.ProjectionType [=] {ALL | KEYS_ONLY | INCLUDE}
    | Projection.NonKeyAttributes [=] ('string', ...)
    | ProvisionedThroughput.ReadCapacityUnits [=] value
    | ProvisionedThroughput.WriteCapacityUnits [=] value
}

replica_option: {
    KMSMasterKeyId [=] value
    | ProvisionedThroughputOverride.ReadCapacityUnits [=] value
    | TableClassOverride [=] {STANDARD | STANDARD_INFREQUENT_ACCESS}
    | GlobalSecondaryIndexes [=] (
        index_name [ProvisionedThroughputOverride.ReadCapacityUnits [=] value],
        ...)
}

table_options:
    table_option [[,] table_option] ...

table_option: {
    BillingMode [=] {PROVISIONED | PAY_PER_REQUEST}
    | ProvisionedThroughput.ReadCapacityUnits [=] value
    | ProvisionedThroughput.WriteCapacityUnits [=] value
    | StreamSpecification.StreamEnabled [=] {True | False}
    | StreamSpecification.StreamViewType [=] {NEW_IMAGE | OLD_IMAGE | NEW_AND_OLD_IMAGES | KEYS_ONLY}
    | SSESpecification.Enabled [=] {True | False}
    | SSESpecification.SSEType [=] {AES256 | KMS}
    | SSESpecification.KMSMasterKeyId [=] value
    | TableClass [=] {STANDARD | STANDARD_INFREQUENT_ACCESS}
}

Sample SQL of Altering Table:
-----------------------------
ALTER TABLE Issues (
    IssueId numeric,
    Title string,
    CreateDate string,
    DueDate string,
    CREATE INDEX CreateDateIndex GLOBAL (
        CreateDate PARTITION KEY,
        IssueId SORT KEY
    )
        Projection.ProjectionType INCLUDE
        Projection.NonKeyAttributes (Description, Status)
        ProvisionedThroughput.ReadCapacityUnits 1
        ProvisionedThroughput.WriteCapacityUnits 1,
    UPDATE INDEX DueDateIndex GLOBAL
        ProvisionedThroughput.ReadCapacityUnits 10
        ProvisionedThroughput.WriteCapacityUnits 10,
    DELETE INDEX IssueIdIndex GLOBAL,
    CREATE REPLICA cn-north-1
        KMSMasterKeyId XXXXXXXX
        ProvisionedThroughputOverride.ReadCapacityUnits 1
        GlobalSecondaryIndexes (
            CreateDateIndex
            ProvisionedThroughputOverride.ReadCapacityUnits 1,
            DueDateIndex
            ProvisionedThroughputOverride.ReadCapacityUnits 1
        )
        TableClassOverride STANDARD,
    UPDATE REPLICA cn-northwest-1
        KMSMasterKeyId *********
        ProvisionedThroughputOverride.ReadCapacityUnits 10
        GlobalSecondaryIndexes (
            CreateDateIndex
            ProvisionedThroughputOverride.ReadCapacityUnits 10,
            DueDateIndex
            ProvisionedThroughputOverride.ReadCapacityUnits 10
        )
        TableClassOverride STANDARD_INFREQUENT_ACCESS,
    DELETE REPLICA cn-northwest-2
)
BillingMode PAY_PER_REQUEST
ProvisionedThroughput.ReadCapacityUnits 10
ProvisionedThroughput.WriteCapacityUnits 10
StreamSpecification.StreamEnabled True
StreamSpecification.StreamViewType NEW_AND_OLD_IMAGES
SSESpecification.Enabled True
SSESpecification.SSEType KMS
SSESpecification.KMSMasterKeyId $$$$$$$$
TableClass STANDARD_INFREQUENT_ACCESS
"""
import logging
from .ddl_sql import DdlBase
from .common import KeyWords, Tokens
from pyparsing import (
    Opt,
    Group,
    ZeroOrMore,
    delimited_list,
    one_of,
    Forward,
    ParseResults,
)
from typing import Any, Dict, Optional

_logger = logging.getLogger(__name__)  # type: ignore


class DdlAlter(DdlBase):
    _ATTRIBUTE = Group(Tokens.ATTRIBUTE_NAME + Tokens.DATA_TYPE)("attribute").set_name(
        "attribute"
    )

    _ALTER_INDEX = Group(
        KeyWords.CREATE("alter_ops") + DdlBase._INDEX
        ^ KeyWords.UPDATE("alter_ops")
        + Group(
            KeyWords.INDEX
            + Tokens.INDEX_NAME
            + Tokens.INDEX_TYPE
            + DdlBase._INDEX_OPTIONS
        )("index").set_name("index")
        ^ KeyWords.DELETE("alter_ops")
        + Group(KeyWords.INDEX + Tokens.INDEX_NAME + Tokens.INDEX_TYPE)(
            "index"
        ).set_name("index")
    )

    _REPLICA = Group(
        KeyWords.REPLICA
        + Tokens.REGION_NAME
        + Opt(
            ZeroOrMore(
                Group(
                    DdlBase._KMSMASTERKEYID
                    + Opt(KeyWords.EQUALS)
                    + Tokens.KMSMASTERKEYID
                    ^ DdlBase._PROVISIONEDTHROUGHPUTOVERRIDE_READCAPACITYUNITS
                    + Opt(KeyWords.EQUALS)
                    + Tokens.INT_VALUE
                    ^ DdlBase._TABLECLASSOVERRIDE
                    + Opt(KeyWords.EQUALS)
                    + one_of("STANDARD STANDARD_INFREQUENT_ACCESS")
                )("replica_option").set_name("replica_option")
                ^ Group(
                    DdlBase._GLOBALSECONDARYINDEXES
                    + KeyWords.LPAR
                    + delimited_list(
                        Group(
                            Tokens.INDEX_NAME
                            + DdlBase._PROVISIONEDTHROUGHPUTOVERRIDE_READCAPACITYUNITS
                            + Opt(KeyWords.EQUALS)
                            + Tokens.INT_VALUE
                        )("gsi").set_name("gsi")
                    )("gsis").set_name("gsis")
                    + KeyWords.RPAR
                )("replica_gsi_option").set_name("replica_gsi_option")
            )("replica_options").set_name("replica_options")
        )
    )("replica").set_name("replica")

    _ALTER_REPLICA = Group(
        KeyWords.CREATE("alter_ops") + _REPLICA
        ^ KeyWords.UPDATE("alter_ops") + _REPLICA
        ^ KeyWords.DELETE("alter_ops")
        + Group(KeyWords.REPLICA + Tokens.REGION_NAME)("replica").set_name("replica")
    )

    _ALTER_TABLE_STATEMENT = (
        KeyWords.ALTER
        + KeyWords.TABLE
        + Tokens.TABLE_NAME
        + KeyWords.LPAR
        + delimited_list(ZeroOrMore(_ATTRIBUTE))("attributes").set_name("attributes")
        + delimited_list(ZeroOrMore(_ALTER_INDEX))("indices").set_name("indices")
        + delimited_list(ZeroOrMore(_ALTER_REPLICA))("replicas").set_name("replicas")
        + KeyWords.RPAR
        + DdlBase._TABLE_OPTIONS
    )("alter_statement").set_name("alter_statement")

    _DDL_ALTER_EXPR = Forward()
    _DDL_ALTER_EXPR <<= _ALTER_TABLE_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)

    @property
    def syntax_def(self) -> Forward:
        return DdlAlter._DDL_ALTER_EXPR

    def transform(self) -> Dict[str, Any]:
        if self.root_parse_results is None:
            raise ValueError("Statement was not parsed yet")

        request = dict()
        attr_def_ = None

        for attr in self.root_parse_results["attributes"]:
            if attr_def_ is None:
                attr_def_ = list()
            attr_name = attr["attribute_name"]
            data_type = attr["data_type"]
            attr_def_.append(self._construct_attr_def(attr_name, data_type))

        if attr_def_ is not None:
            request.update({"AttributeDefinitions": attr_def_})

        table_name_ = self.root_parse_results["table"]
        request.update({"TableName": table_name_})

        gsis_ = None
        indices = self.root_parse_results["indices"]
        if indices is not None:
            for index in indices:
                alter_ops = index["alter_ops"]
                index_type = index["index"]["index_type"]

                if index_type == "GLOBAL":
                    if gsis_ is None:
                        gsis_ = list()

                    if alter_ops == "CREATE" or alter_ops == "UPDATE":
                        gsis_.append(
                            {
                                self._get_alter_operation(
                                    alter_ops
                                ): self._construct_index(index["index"])
                            }
                        )
                    elif alter_ops == "DELETE":
                        gsis_.append(
                            {
                                self._get_alter_operation(alter_ops): {
                                    "IndexName": index["index"]["index_name"]
                                }
                            }
                        )
                else:
                    raise LookupError("Index type is invalid")

        if gsis_ is not None:
            request.update({"GlobalSecondaryIndexUpdates": gsis_})

        replicas_ = None
        replicas = self.root_parse_results["replicas"]
        if replicas is not None:
            for replica in replicas:
                if replicas_ is None:
                    replicas_ = list()
                alter_ops = replica["alter_ops"]

                if alter_ops == "CREATE" or alter_ops == "UPDATE":
                    replicas_.append(
                        {
                            self._get_alter_operation(
                                alter_ops
                            ): self._construct_replica(replica["replica"])
                        }
                    )
                elif alter_ops == "DELETE":
                    replicas_.append(
                        {
                            self._get_alter_operation(alter_ops): {
                                "RegionName": replica["replica"]["region_name"]
                            }
                        }
                    )

        if replicas_ is not None:
            request.update({"ReplicaUpdates": replicas_})

        table_options_ = self.root_parse_results["table_options"]
        if table_options_ is not None:
            converted_options = self._construct_options(table_options_)
            if converted_options is not None:
                request.update(converted_options)

        return request

    def _get_alter_operation(self, ops: str) -> Optional[str]:
        if ops is None:
            return None

        if ops.lower() == "CREATE".lower():
            return "Create"
        elif ops.lower() == "UPDATE".lower():
            return "Update"
        elif ops.lower() == "DELETE".lower():
            return "Delete"
        else:
            raise LookupError("Alter operation is invalid")

    def _construct_replica(self, replica: ParseResults) -> Dict[str, Any]:
        replicas_ = dict()

        region_name = replica["region_name"]
        replicas_.update({"RegionName": region_name})

        replica_options = replica.get("replica_options")
        replica_options_ = None
        replica_gsi_option_ = None

        if replica_options is not None:
            for replica_option in replica_options:
                if replica_options_ is None:
                    replica_options_ = list()

                if replica_option.get_name() == "replica_gsi_option":
                    replica_gsi_option_ = replica_option
                else:
                    replica_options_.append(replica_option)

        if replica_options_ is not None:
            converted_options_ = self._construct_options(replica_options_)
            if converted_options_ is not None:
                replicas_.update(converted_options_)

        if replica_gsi_option_ is not None:
            replicas_.update(
                {
                    "GlobalSecondaryIndexes": self._construct_replica_gsi(
                        replica_gsi_option_
                    )
                }
            )

        return replicas_

    def _construct_replica_gsi(
        self, replica_gsi_option: ParseResults
    ) -> Optional[Dict[str, Any]]:
        gsis = replica_gsi_option.get("gsis")
        gsis_ = None

        if gsis is not None:
            for gsi in gsis:
                index_name = gsi["index_name"]

                if len(gsi) == 3:
                    if gsis_ is None:
                        gsis_ = list()

                    option_path_name = gsi[1]
                    option_value = gsi[2]
                    options = self._parse_option_path(option_path_name, option_value)
                    converted_ = {"IndexName": index_name}

                    if options is not None and len(options) > 0:
                        converted_.update(options)
                    gsis_.append(converted_)

        return gsis_



# -*- coding: utf-8 -*-
import logging
from typing import List, Optional, Any, Dict
from collections import OrderedDict
from .sql.dml_sql import DmlFunction
from .sql.base import Base
from .sql.common import QueryType
from .sql.parser import SQLParser
from .error import OperationalError

_logger = logging.getLogger(__name__)  # type: ignore


class ColumnInfo:
    def __init__(
        self,
        name: str,
        original: str,
        alias: str = None,
        function: DmlFunction = None,
        type_code: str = None,
        display_size: str = None,
        internal_size: str = None,
        precision: str = None,
        scale: str = None,
        null_ok: str = None,
    ) -> None:
        self._name = name
        self._original = original
        self._alias = alias
        self._function = function
        self._type_code = type_code
        self._display_size = display_size
        self._internal_size = internal_size
        self._precision = precision
        self._scale = scale
        self._null_ok = null_ok

    @property
    def name(self) -> str:
        return self._name

    @property
    def original(self) -> str:
        return self._original

    @property
    def alias(self) -> str:
        return self._alias

    @alias.setter
    def alias(self, value: str) -> None:
        self._alias = value

    @property
    def function(self) -> DmlFunction:
        return self._function

    @property
    def type_code(self) -> str:
        return self._type_code

    @type_code.setter
    def type_code(self, value: str) -> None:
        self._type_code = value

    @property
    def display_size(self) -> str:
        return self._display_size

    @display_size.setter
    def display_size(self, value: str) -> None:
        self._display_size = value

    @property
    def internal_size(self) -> str:
        return self._internal_size

    @internal_size.setter
    def internal_size(self, value: str) -> None:
        self._internal_size = value

    @property
    def precision(self) -> str:
        return self._precision

    @precision.setter
    def precision(self, value: str) -> None:
        self._precision = value

    @property
    def scale(self) -> str:
        return self._scale

    @scale.setter
    def scale(self, value: str) -> None:
        self._scale = value

    @property
    def null_ok(self) -> str:
        return self._null_ok

    @null_ok.setter
    def null_ok(self, value: str) -> None:
        self._null_ok = value

    def __str__(self) -> str:
        return "%s | %s | %s | %s | %s | %s | %s | %s | %s" % (
            self.name,
            self.original,
            self.alias,
            self.type_code,
            self.display_size,
            self.internal_size,
            self.precision,
            self.scale,
            self.null_ok,
        )


class Metadata:
    def __init__(self, column_infos: List[Optional[ColumnInfo]] = None) -> None:
        self._column_infos = OrderedDict()

        if column_infos is not None:
            for column_info in column_infos:
                self.append(column_info)

    def get(self, name: str, default: Any = None) -> ColumnInfo:
        return self._column_infos.get(name, default)

    def append(self, column_info: ColumnInfo) -> None:
        self.update(column_info)

    def update(self, column_info: ColumnInfo) -> None:
        self._column_infos.update({column_info.name: column_info})

    def clear(self) -> None:
        self._column_infos.clear()

    def index(self, key: str) -> int:
        return list(self._column_infos).index(key)

    def keys(self) -> List[str]:
        return self._column_infos.keys()

    def __len__(self):
        return len(self._column_infos.keys())

    def __getitem__(self, key: str) -> ColumnInfo:
        return self._column_infos[key]

    def __contains__(self, key: str):
        return key in self._column_infos

    def __next__(self):
        keys = list(self._column_infos)
        if self.__iterator_index < len(keys):
            ret = self._column_infos[keys[self.__iterator_index]]
            self.__iterator_index += 1
            return ret
        raise StopIteration

    def __iter__(self):
        self.__iterator_index = 0
        return self


class Statement:
    def __init__(self, operation: str, parser_class: Base = None) -> None:
        self._sql_parser = SQLParser(operation, parser_class=parser_class)
        self._api_request = self._sql_parser.transform()

    @property
    def sql_parser(self) -> SQLParser:
        return self._sql_parser

    @property
    def api_request(self) -> Optional[Dict[str, Any]]:
        return self._api_request


class Statements:
    def __init__(self, statements: List[Statement] = None) -> None:
        if statements is not None:
            for statement_ in statements:
                self._validate(statement_)
            self._statments = statements
        else:
            self._statments = list()

        self._query_type = None
        self._query_category = None
        self._operation_type = None
        self._limit = None

    @property
    def query_type(self) -> QueryType:
        return self._query_type

    @property
    def query_category(self) -> str:
        return self._query_category

    @property
    def operation_type(self) -> str:
        return self._operation_type

    @property
    def limit(self) -> str:
        return self._limit

    def _validate(self, statement: Statement) -> bool:
        if statement is not None:
            query_type_ = statement.sql_parser.query_type
            query_category_ = statement.sql_parser.query_category
            operation_type_ = statement.sql_parser.operation_type
            if self._query_type is None:
                self._query_type = query_type_
                self._query_category = query_category_
                self._operation_type = operation_type_

                if (
                    self._query_type == QueryType.SELECT
                    or self._query_type == QueryType.LIST
                ):
                    self._limit = statement.sql_parser.parser.limit

                return True
            else:
                if query_category_ == "DML" and operation_type_ == self._query_type[1]:
                    return True
                elif query_category_ == "DDL" or query_category_ == "UTIL":
                    raise OperationalError("Not support batch DDL/Utility statements")
                else:
                    raise OperationalError(
                        "Not support both read and write operation statements in the request"
                    )

    def append(self, statement: Statement) -> None:
        if self._validate(statement):
            self._statments.append(statement)

    def extend(self, statements: "Statements") -> None:
        if statements is not None:
            for statement_ in statements:
                self.append(statement_)

    def clear(self) -> None:
        self._statments.clear()
        self._query_type = None

    def __len__(self):
        return len(self._statments)

    def __getitem__(self, key):
        return self._statments[key]

    def __next__(self):
        if self.__iterator_index < len(self._statments):
            ret = self._statments[self.__iterator_index]
            self.__iterator_index += 1
            return ret
        raise StopIteration

    def __iter__(self):
        self.__iterator_index = 0
        return self



# -*- coding: utf-8 -*-
import functools
import logging
import threading
from collections.abc import MutableMapping, MutableSequence
from typing import Any, Callable, Iterable, Dict

import tenacity
from tenacity import after_log, retry_if_exception, stop_after_attempt, wait_exponential

_logger = logging.getLogger(__name__)  # type: ignore


def synchronized(wrapped: Callable[..., Any]) -> Any:
    """The missing @synchronized decorator

    https://git.io/vydTA"""
    _lock = threading.RLock()

    @functools.wraps(wrapped)
    def _wrapper(*args, **kwargs):
        with _lock:
            return wrapped(*args, **kwargs)

    return _wrapper


# Copy from distutils.util to remove dependency
def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError("invalid truth value %r" % (val,))


def flatten_dict(
    d: Dict[str, Any], parent_key: str = "", separator: str = "."
) -> Dict[str, Any]:
    items = list()
    for key, val in d.items():
        new_key = parent_key + separator + key if parent_key else key
        if isinstance(val, MutableMapping):
            items.extend(flatten_dict(val, new_key, separator=separator).items())
        elif isinstance(val, MutableSequence):
            for i, v in enumerate(val):
                new_list_key = "%s[%s]" % (new_key, str(i))
                if isinstance(v, MutableMapping):
                    items.extend(
                        flatten_dict(v, new_list_key, separator=separator).items()
                    )
                else:
                    items.append((new_list_key, v))
        else:
            items.append((new_key, val))
    return dict(items)


class RetryConfig:
    def __init__(
        self,
        exceptions: Iterable[str] = (
            "ThrottlingException",
            "TooManyRequestsException",
        ),
        attempt: int = 5,
        multiplier: int = 1,
        max_delay: int = 100,
        exponential_base: int = 2,
    ) -> None:
        self.exceptions = exceptions
        self.attempt = attempt
        self.multiplier = multiplier
        self.max_delay = max_delay
        self.exponential_base = exponential_base


def retry_api_call(
    func: Callable[..., Any],
    config: RetryConfig,
    logger: logging.Logger = None,
    *args,
    **kwargs,
) -> Any:
    retry = tenacity.Retrying(
        retry=retry_if_exception(
            lambda e: (
                getattr(e, "response", {}).get("Error", {}).get("Code", None)
                in config.exceptions
                if e
                else False
            )
        ),
        stop=stop_after_attempt(config.attempt),
        wait=wait_exponential(
            multiplier=config.multiplier,
            max=config.max_delay,
            exp_base=config.exponential_base,
        ),
        after=after_log(logger, logger.level) if logger else None,  # type: ignore
        reraise=True,
    )
    return retry(func, *args, **kwargs)



# -*- coding: utf-8 -*-
import logging
from typing import TYPE_CHECKING, Dict, Any, List, Optional

from .helper import QueryDBHelper
from .querydb import QueryDB
from .dml_select import SupersetSelect
from ..converter import Converter
from ..model import Statements, Statement, ColumnInfo
from ..util import RetryConfig, synchronized
from ..cursor import Cursor
from ..result_set import DynamoDBResultSet
from ..executor import BaseExecutor, DmlStatementExecutor
from ..error import OperationalError

if TYPE_CHECKING:
    from ..connection import Connection

_logger = logging.getLogger(__name__)  # type: ignore


class SupersetCursor(Cursor):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._result_set_class = SupersetResultSet

    @synchronized
    def execute(
        self: Cursor, operation: str, parameters: Optional[List[Dict[str, Any]]] = None
    ) -> Cursor:
        statement = Statement(operation, SupersetSelect)
        return self.execute_statement(statement, parameters)


class SupersetResultSet(DynamoDBResultSet):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        arraysize: int,
        retry_config: RetryConfig,
        is_transaction: bool = False,
        executor_class: BaseExecutor = None,
        **kwargs
    ) -> None:
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            arraysize=arraysize,
            retry_config=retry_config,
            is_transaction=is_transaction,
            executor_class=SupersetStatementExecutor,
            **kwargs
        )


class SupersetStatementExecutor(DmlStatementExecutor):
    def __init__(
        self,
        connection: "Connection",
        converter: Converter,
        statements: Statements,
        retry_config: RetryConfig,
        **kwargs
    ) -> None:
        self._superset_table: str = "SUPERSET_QUERY"
        self._statement: Statement = statements[0]
        self._query_db: QueryDB = QueryDBHelper.create(self._statement, **kwargs)
        self._querydb_load_batch_size: int = self._query_db.config.load_batch_size
        super().__init__(
            connection=connection,
            converter=converter,
            statements=statements,
            retry_config=retry_config,
            **kwargs
        )

    def pre_execute(self) -> None:
        self.execute()

    def execute(self, **kwargs) -> None:
        try:
            parser = self._statement.sql_parser.parser
            if parser.is_nested:
                if not self._query_db.has_cache():
                    self._query_db.init_cache_table()
                    self._query_db.drop_query_table()
                    with self.connection.cursor() as cursor:
                        cursor.result_set_class = DynamoDBResultSet
                        cursor.execute_statement(self._statement)
                        self._load_into_query_db(cursor.result_set)

                (desc_, results_) = self._query_db.query()
                self._rows.extend(results_)
                for d in desc_:
                    self._metadata.update(
                        ColumnInfo(
                            d[0],
                            d[0],
                            None,
                            None,
                            d[1],
                        )
                    )

            else:
                super().execute(**kwargs)
        except Exception as e:
            self._query_db.rollback()
            _logger.exception("Failed to execute statement.")
            raise OperationalError(*e.args) from e
        finally:
            self._query_db.close()

    def _load_into_query_db(self, ddb_result_set: DynamoDBResultSet) -> None:
        self._query_db.create_query_table(ddb_result_set.metadata)

        raw_data = ddb_result_set.fetchmany(self._querydb_load_batch_size)
        while True:
            if len(raw_data) > 0:
                self._query_db.write_raw_data(ddb_result_set.metadata, raw_data)
                raw_data = ddb_result_set.fetchmany(self._querydb_load_batch_size)
            else:
                break



# -*- coding: utf-8 -*-
import logging
from abc import ABCMeta, abstractmethod
from pyparsing import ParseResults, Forward
from typing import Any, Dict

_logger = logging.getLogger(__name__)  # type: ignore


class Base(metaclass=ABCMeta):
    def __init__(self, statement: str) -> None:
        self._statement = statement
        self._executed_statement = statement
        self._root_parse_results = None

        self.preprocess()

    @property
    def statement(self):
        return self._statement

    @property
    def executed_statement(self):
        return self._executed_statement

    @property
    def root_parse_results(self) -> ParseResults:
        if self._root_parse_results is not None:
            return self._root_parse_results

        if self._statement is None:
            raise ValueError("Statement is not specified")

        self._root_parse_results = self.syntax_def.parseString(self._executed_statement)
        return self._root_parse_results

    @abstractmethod
    def syntax_def(self) -> Forward:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def transform(self) -> Dict[str, Any]:
        raise NotImplementedError  # pragma: no cover

    def preprocess(self) -> None:
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
"""
Syntax of PartiQL
https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/ql-reference.insert.html
------------------
INSERT INTO table VALUE item

Sample SQL of Inserting:
------------------------
INSERT INTO "Music" value {'Artist' : 'Acme Band','SongTitle' : 'PartiQL Rocks'}
"""
import logging
from .dml_sql import DmlBase
from .common import KeyWords, Tokens
from pyparsing import Forward, Group, Regex
from typing import Any, Dict

_logger = logging.getLogger(__name__)  # type: ignore


class DmlInsert(DmlBase):

    # Parse everything inside curly brackets as a single string
    _ITEM = Group(
        KeyWords.LCURLYBRACKET
        + Regex(r"[^}]*")("item_content")
        + KeyWords.RCURLYBRACKET
    )("item").set_name("item")

    _INSERT_STATEMENT = (
        KeyWords.INSERT + KeyWords.INTO + Tokens.TABLE_NAME + KeyWords.VALUE + _ITEM
    )("insert_statement").set_name("insert_statement")

    _DML_INSERT_EXPR = Forward()
    _DML_INSERT_EXPR <<= _INSERT_STATEMENT

    def __init__(self, statement: str) -> None:
        super().__init__(statement)

    @property
    def syntax_def(self) -> Forward:
        return DmlInsert._DML_INSERT_EXPR

    def transform(self) -> Dict[str, Any]:
        table_name_ = self.root_parse_results["table"]
        item_content_ = self.root_parse_results["item"]["item_content"]

        table_ = '"%s"' % table_name_
        item_ = "{%s}" % item_content_

        statement_ = "INSERT INTO {table} VALUE {item}"
        statement_ = statement_.format(
            table=table_,
            item=item_,
        )
        request = {"Statement": statement_.strip()}

        return request



# -*- coding: utf-8 -*-
import re
from ..util import strtobool
from typing import TYPE_CHECKING, cast, Optional, Sequence, List

import pydynamodb
from pydynamodb.error import OperationalError, NotSupportedError
from ..sql.common import RESERVED_WORDS

import botocore
from sqlalchemy import exc, types, util, __version__
from sqlalchemy.engine import Engine, reflection
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.sql import crud
from sqlalchemy.sql.compiler import (
    IdentifierPreparer,
    DDLCompiler,
    GenericTypeCompiler,
    SQLCompiler,
)

if TYPE_CHECKING:
    from types import ModuleType


def _check_sqla_major_ver():  # pragma: no cover
    try:
        return tuple(int(x) for x in __version__.split("."))[0]
    except ValueError | Exception:
        raise NotSupportedError("Invalid SQLAlchemy version: %s" % __version__)


_SQLALCHEMY_MAJOR_VERSION = _check_sqla_major_ver()

if _SQLALCHEMY_MAJOR_VERSION > 2:
    from sqlalchemy.sql.compiler import InsertmanyvaluesSentinelOpts, _InsertManyValues


class DynamoDBIdentifierPreparer(IdentifierPreparer):
    reserved_words = RESERVED_WORDS


class DynamoDBDDLCompiler(DDLCompiler):
    def __init__(
        self,
        dialect,
        statement,
        schema_translate_map=None,
        compile_kwargs=None,
    ):
        exc.CompileError("DDL statement is not supported by DDB PartiQL.")


class DynamoDBStatementCompiler(SQLCompiler):
    # DynamoDB can't guarantee the column orders of result
    # _textual_ordered_columns: bool = True
    _ordered_columns: bool = False

    def visit_column(
        self,
        column,
        add_to_result_map=None,
        include_table: bool = True,
        result_map_targets=(),
        ambiguous_table_name_map=None,
        **kwargs,
    ) -> str:
        return super().visit_column(
            column=column,
            add_to_result_map=add_to_result_map,
            include_table=False,
            result_map_targets=result_map_targets,
            ambiguous_table_name_map=ambiguous_table_name_map,
            kwargs=kwargs,
        )

    def visit_label(
        self,
        label,
        add_to_result_map=None,
        within_label_clause=False,
        within_columns_clause=False,
        render_label_as_label=None,
        result_map_targets=(),
        **kw,
    ):
        return super().visit_label(
            label=label,
            add_to_result_map=add_to_result_map,
            within_label_clause=within_label_clause,
            within_columns_clause=False,
            render_label_as_label=render_label_as_label,
            result_map_targets=result_map_targets,
            kw=kw,
        )

    def limit_clause(self, select, **kw):
        if hasattr(select, "_simple_int_clause"):
            limit_clause = select._limit_clause
            if limit_clause is not None and select._simple_int_clause(limit_clause):
                return f" LIMIT {self.process(limit_clause.render_literal_execute(), **kw)}"
        else:
            if select._limit_clause is not None:
                return " LIMIT " + self.process(select._limit_clause, **kw)
        return ""

    def visit_insert(self, insert_stmt, **kw):  # pragma: no cover
        # Compatible design for SQLAlchemy v1 and v2
        if _SQLALCHEMY_MAJOR_VERSION < 2:
            return self._visit_insert_v1(insert_stmt, **kw)
        return self._visit_insert_v2(insert_stmt, None, None, **kw)

    def _visit_insert_v1(self, insert_stmt, **kw):  # pragma: no cover
        compile_state = insert_stmt._compile_state_factory(insert_stmt, self, **kw)
        insert_stmt = compile_state.statement

        toplevel = not self.stack

        if toplevel:
            self.isinsert = True
            if not self.dml_compile_state:
                self.dml_compile_state = compile_state
            if not self.compile_state:
                self.compile_state = compile_state

        self.stack.append(
            {
                "correlate_froms": set(),
                "asfrom_froms": set(),
                "selectable": insert_stmt,
            }
        )

        crud_params = crud._get_crud_params(self, insert_stmt, compile_state, **kw)

        if (
            not crud_params
            and not self.dialect.supports_default_values
            and not self.dialect.supports_default_metavalue
            and not self.dialect.supports_empty_insert
        ):
            raise exc.CompileError(
                "The '%s' dialect with current database "
                "version settings does not support empty "
                "inserts." % self.dialect.name
            )

        preparer = self.preparer

        text = "INSERT "

        if insert_stmt._prefixes:
            text += self._generate_prefixes(insert_stmt, insert_stmt._prefixes, **kw)

        text += "INTO "
        table_text = preparer.format_table(insert_stmt.table)

        if insert_stmt._hints:
            _, table_text = self._setup_crud_hints(insert_stmt, table_text)

        if insert_stmt._independent_ctes:
            for cte in insert_stmt._independent_ctes:
                cte._compiler_dispatch(self, **kw)

        text += table_text

        if self.returning or insert_stmt._returning:
            returning_clause = self.returning_clause(
                insert_stmt, self.returning or insert_stmt._returning
            )

            if self.returning_precedes_values:
                text += " " + returning_clause
        else:
            returning_clause = None

        if insert_stmt.select is not None:
            raise exc.CompileError(
                "The '%s' dialect with current database "
                "version settings does not support "
                "select statement in insert." % self.dialect.name
            )
        else:
            insert_single_values_expr = ", ".join(
                ["'%s': %s" % (expr, value) for c, expr, value in crud_params]
            )
            text += " VALUE {%s}" % insert_single_values_expr
            if toplevel:
                self.insert_single_values_expr = insert_single_values_expr

        if insert_stmt._post_values_clause is not None:
            post_values_clause = self.process(insert_stmt._post_values_clause, **kw)
            if post_values_clause:
                text += " " + post_values_clause

        if returning_clause and not self.returning_precedes_values:
            text += " " + returning_clause

        if self.ctes and not self.dialect.cte_follows_insert:
            nesting_level = len(self.stack) if not toplevel else None
            text = (
                self._render_cte_clause(
                    nesting_level=nesting_level, include_following_stack=True
                )
                + text
            )

        self.stack.pop(-1)

        return text

    def _visit_insert_v2(
        self, insert_stmt, visited_bindparam=None, visiting_cte=None, **kw
    ):  # pragma: no cover
        compile_state = insert_stmt._compile_state_factory(insert_stmt, self, **kw)
        insert_stmt = compile_state.statement

        if visiting_cte is not None:
            kw["visiting_cte"] = visiting_cte
            toplevel = False
        else:
            toplevel = not self.stack

        if toplevel:
            self.isinsert = True
            if not self.dml_compile_state:
                self.dml_compile_state = compile_state
            if not self.compile_state:
                self.compile_state = compile_state

        self.stack.append(
            {
                "correlate_froms": set(),
                "asfrom_froms": set(),
                "selectable": insert_stmt,
            }
        )

        counted_bindparam = 0

        # reset any incoming "visited_bindparam" collection
        visited_bindparam = None

        # for positional, insertmanyvalues needs to know how many
        # bound parameters are in the VALUES sequence; there's no simple
        # rule because default expressions etc. can have zero or more
        # params inside them.   After multiple attempts to figure this out,
        # this very simplistic "count after" works and is
        # likely the least amount of callcounts, though looks clumsy
        if self.positional and visiting_cte is None:
            # if we are inside a CTE, don't count parameters
            # here since they wont be for insertmanyvalues. keep
            # visited_bindparam at None so no counting happens.
            # see #9173
            visited_bindparam = []

        crud_params_struct = crud._get_crud_params(
            self,
            insert_stmt,
            compile_state,
            toplevel,
            visited_bindparam=visited_bindparam,
            **kw,
        )

        if self.positional and visited_bindparam is not None:
            counted_bindparam = len(visited_bindparam)
            if self._numeric_binds:
                if self._values_bindparam is not None:
                    self._values_bindparam += visited_bindparam
                else:
                    self._values_bindparam = visited_bindparam

        crud_params_single = crud_params_struct.single_params

        if (
            not crud_params_single
            and not self.dialect.supports_default_values
            and not self.dialect.supports_default_metavalue
            and not self.dialect.supports_empty_insert
        ):
            raise exc.CompileError(
                "The '%s' dialect with current database "
                "version settings does not support empty "
                "inserts." % self.dialect.name
            )

        if compile_state._has_multi_parameters:
            if not self.dialect.supports_multivalues_insert:
                raise exc.CompileError(
                    "The '%s' dialect with current database "
                    "version settings does not support "
                    "in-place multirow inserts." % self.dialect.name
                )
            elif (
                self.implicit_returning or insert_stmt._returning
            ) and insert_stmt._sort_by_parameter_order:
                raise exc.CompileError(
                    "RETURNING cannot be determinstically sorted when "
                    "using an INSERT which includes multi-row values()."
                )
            crud_params_single = crud_params_struct.single_params
        else:
            crud_params_single = crud_params_struct.single_params

        preparer = self.preparer
        supports_default_values = self.dialect.supports_default_values

        text = "INSERT "

        if insert_stmt._prefixes:
            text += self._generate_prefixes(insert_stmt, insert_stmt._prefixes, **kw)

        text += "INTO "
        table_text = preparer.format_table(insert_stmt.table)

        if insert_stmt._hints:
            _, table_text = self._setup_crud_hints(insert_stmt, table_text)

        if insert_stmt._independent_ctes:
            self._dispatch_independent_ctes(insert_stmt, kw)

        text += table_text

        # look for insertmanyvalues attributes that would have been configured
        # by crud.py as it scanned through the columns to be part of the
        # INSERT
        use_insertmanyvalues = crud_params_struct.use_insertmanyvalues
        named_sentinel_params: Optional[Sequence[str]] = None
        add_sentinel_cols = None
        implicit_sentinel = False

        returning_cols = self.implicit_returning or insert_stmt._returning
        if returning_cols:
            add_sentinel_cols = crud_params_struct.use_sentinel_columns
            if add_sentinel_cols is not None:
                assert use_insertmanyvalues

                # search for the sentinel column explicitly present
                # in the INSERT columns list, and additionally check that
                # this column has a bound parameter name set up that's in the
                # parameter list.  If both of these cases are present, it means
                # we will have a client side value for the sentinel in each
                # parameter set.

                _params_by_col = {
                    col: param_names for col, _, _, param_names in crud_params_single
                }
                named_sentinel_params = []
                for _add_sentinel_col in add_sentinel_cols:
                    if _add_sentinel_col not in _params_by_col:
                        named_sentinel_params = None
                        break
                    param_name = self._within_exec_param_key_getter(_add_sentinel_col)
                    if param_name not in _params_by_col[_add_sentinel_col]:
                        named_sentinel_params = None
                        break
                    named_sentinel_params.append(param_name)

                if named_sentinel_params is None:
                    # if we are not going to have a client side value for
                    # the sentinel in the parameter set, that means it's
                    # an autoincrement, an IDENTITY, or a server-side SQL
                    # expression like nextval('seqname').  So this is
                    # an "implicit" sentinel; we will look for it in
                    # RETURNING
                    # only, and then sort on it.  For this case on PG,
                    # SQL Server we have to use a special INSERT form
                    # that guarantees the server side function lines up with
                    # the entries in the VALUES.
                    if (
                        self.dialect.insertmanyvalues_implicit_sentinel
                        & InsertmanyvaluesSentinelOpts.ANY_AUTOINCREMENT
                    ):
                        implicit_sentinel = True
                    else:
                        # here, we are not using a sentinel at all
                        # and we are likely the SQLite dialect.
                        # The first add_sentinel_col that we have should not
                        # be marked as "insert_sentinel=True".  if it was,
                        # an error should have been raised in
                        # _get_sentinel_column_for_table.
                        assert not add_sentinel_cols[0]._insert_sentinel, (
                            "sentinel selection rules should have prevented "
                            "us from getting here for this dialect"
                        )

                # always put the sentinel columns last.  even if they are
                # in the returning list already, they will be there twice
                # then.
                returning_cols = list(returning_cols) + list(add_sentinel_cols)

            returning_clause = self.returning_clause(
                insert_stmt,
                returning_cols,
                populate_result_map=toplevel,
            )

            if self.returning_precedes_values:
                text += " " + returning_clause

        else:
            returning_clause = None

        if insert_stmt.select is not None:
            # placed here by crud.py
            select_text = self.process(
                self.stack[-1]["insert_from_select"], insert_into=True, **kw
            )

            if self.ctes and self.dialect.cte_follows_insert:
                nesting_level = len(self.stack) if not toplevel else None
                text += " %s%s" % (
                    self._render_cte_clause(
                        nesting_level=nesting_level,
                        include_following_stack=True,
                    ),
                    select_text,
                )
            else:
                text += " %s" % select_text
        elif not crud_params_single and supports_default_values:
            text += " DEFAULT VALUES"
            if use_insertmanyvalues:
                self._insertmanyvalues = _InsertManyValues(
                    True,
                    self.dialect.default_metavalue_token,
                    cast("List[crud._CrudParamElementStr]", crud_params_single),
                    counted_bindparam,
                    sort_by_parameter_order=(insert_stmt._sort_by_parameter_order),
                    includes_upsert_behaviors=(
                        insert_stmt._post_values_clause is not None
                    ),
                    sentinel_columns=add_sentinel_cols,
                    num_sentinel_columns=(
                        len(add_sentinel_cols) if add_sentinel_cols else 0
                    ),
                    implicit_sentinel=implicit_sentinel,
                )
        elif compile_state._has_multi_parameters:
            text += " VALUES %s" % (
                ", ".join(
                    "(%s)" % (", ".join(value for _, _, value, _ in crud_param_set))
                    for crud_param_set in crud_params_struct.all_multi_params
                ),
            )
        else:
            insert_single_values_expr = ", ".join(
                [
                    "'%s': %s" % (expr, value)
                    for _, expr, value, _ in cast(
                        "List[crud._CrudParamElementStr]",
                        crud_params_single,
                    )
                ]
            )

            if use_insertmanyvalues:
                if (
                    implicit_sentinel
                    and (
                        self.dialect.insertmanyvalues_implicit_sentinel
                        & InsertmanyvaluesSentinelOpts.USE_INSERT_FROM_SELECT
                    )
                    # this is checking if we have
                    # INSERT INTO table (id) VALUES (DEFAULT).
                    and not (crud_params_struct.is_default_metavalue_only)
                ):
                    # if we have a sentinel column that is server generated,
                    # then for selected backends render the VALUES list as a
                    # subquery.  This is the orderable form supported by
                    # PostgreSQL and SQL Server.
                    embed_sentinel_value = True

                    render_bind_casts = (
                        self.dialect.insertmanyvalues_implicit_sentinel
                        & InsertmanyvaluesSentinelOpts.RENDER_SELECT_COL_CASTS
                    )

                    colnames = ", ".join(
                        f"p{i}" for i, _ in enumerate(crud_params_single)
                    )

                    if render_bind_casts:
                        # render casts for the SELECT list.  For PG, we are
                        # already rendering bind casts in the parameter list,
                        # selectively for the more "tricky" types like ARRAY.
                        # however, even for the "easy" types, if the parameter
                        # is NULL for every entry, PG gives up and says
                        # "it must be TEXT", which fails for other easy types
                        # like ints.  So we cast on this side too.
                        colnames_w_cast = ", ".join(
                            self.render_bind_cast(
                                col.type,
                                col.type._unwrapped_dialect_impl(self.dialect),
                                f"p{i}",
                            )
                            for i, (col, *_) in enumerate(crud_params_single)
                        )
                    else:
                        colnames_w_cast = colnames

                    text += (
                        f" SELECT {colnames_w_cast} FROM "
                        f"(VALUES ({insert_single_values_expr})) "
                        f"AS imp_sen({colnames}, sen_counter) "
                        "ORDER BY sen_counter"
                    )
                else:
                    # otherwise, if no sentinel or backend doesn't support
                    # orderable subquery form, use a plain VALUES list
                    embed_sentinel_value = False
                    text += f" VALUE {{{insert_single_values_expr}}}"

                self._insertmanyvalues = _InsertManyValues(
                    is_default_expr=False,
                    single_values_expr=insert_single_values_expr,
                    insert_crud_params=cast(
                        "List[crud._CrudParamElementStr]",
                        crud_params_single,
                    ),
                    num_positional_params_counted=counted_bindparam,
                    sort_by_parameter_order=(insert_stmt._sort_by_parameter_order),
                    includes_upsert_behaviors=(
                        insert_stmt._post_values_clause is not None
                    ),
                    sentinel_columns=add_sentinel_cols,
                    num_sentinel_columns=(
                        len(add_sentinel_cols) if add_sentinel_cols else 0
                    ),
                    sentinel_param_keys=named_sentinel_params,
                    implicit_sentinel=implicit_sentinel,
                    embed_values_counter=embed_sentinel_value,
                )

            else:
                text += f" VALUE {{{insert_single_values_expr}}}"

        if insert_stmt._post_values_clause is not None:
            post_values_clause = self.process(insert_stmt._post_values_clause, **kw)
            if post_values_clause:
                text += " " + post_values_clause

        if returning_clause and not self.returning_precedes_values:
            text += " " + returning_clause

        if self.ctes and not self.dialect.cte_follows_insert:
            nesting_level = len(self.stack) if not toplevel else None
            text = (
                self._render_cte_clause(
                    nesting_level=nesting_level,
                    include_following_stack=True,
                )
                + text
            )

        self.stack.pop(-1)

        return text

    def visit_update(self, update_stmt, **kw):
        return super().visit_update(update_stmt, **kw)

    def visit_delete(self, delete_stmt, **kw):
        return super().visit_delete(delete_stmt, **kw)


class DynamoDBTypeCompiler(GenericTypeCompiler):

    def visit_FLOAT(self, type_, **kw):
        return self.visit_REAL(type_, **kw)

    def visit_REAL(self, type_, **kw):
        return "DOUBLE"

    def visit_NUMERIC(self, type_, **kw):
        return self.visit_REAL(type_, **kw)

    def visit_DECIMAL(self, type_, **kw):
        return self.visit_REAL(type_, **kw)

    def visit_INTEGER(self, type_, **kw):
        return "INTEGER"

    def visit_SMALLINT(self, type_, **kw):
        return "INTEGER"

    def visit_BIGINT(self, type_, **kw):
        return "INTEGER"

    def visit_CLOB(self, type_, **kw):
        return self.visit_BINARY(type_, **kw)

    def visit_NCLOB(self, type_, **kw):
        return self.visit_BINARY(type_, **kw)

    def visit_JSON(self, type_, **kw):
        return "JSON"


class DynamoDBDialect(DefaultDialect):
    name = "dynamodb"
    driver = ""
    preparer = DynamoDBIdentifierPreparer
    statement_compiler = DynamoDBStatementCompiler
    ddl_compiler = DynamoDBDDLCompiler
    type_compiler = DynamoDBTypeCompiler
    default_paramstyle = pydynamodb.paramstyle
    supports_sane_rowcount = False
    supports_alter = False
    supports_pk_autoincrement = False
    supports_default_values = False
    supports_empty_insert = False
    supports_multivalues_insert = False
    supports_native_decimal = False
    supports_native_boolean = True
    supports_unicode_statements = True
    supports_unicode_binds = True
    supports_statement_cache = True
    returns_unicode_strings = True
    description_encoding = None
    postfetch_lastrowid = False

    _connect_options = dict()  # type: ignore

    @classmethod
    def dbapi(cls):
        return pydynamodb

    @classmethod
    def import_dbapi(cls) -> "ModuleType":
        return pydynamodb

    def _raw_connection(self, connection):
        if isinstance(connection, Engine):
            return connection.raw_connection()
        return connection.connection

    def create_connect_args(self, url):
        # Connection string format:
        #   dynamodb://
        #   {aws_access_key_id}:{aws_secret_access_key}@dynamodb.{region_name}.amazonaws.com:443
        #   ?verify=false&...
        self._connect_options = self._create_connect_args(url)
        return [[], self._connect_options]

    def _create_connect_args(self, url):
        opts = {
            "aws_access_key_id": url.username if url.username else None,
            "aws_secret_access_key": url.password if url.password else None,
            "region_name": re.sub(
                r"^dynamodb\.([a-z0-9-]+)\.amazonaws\.(com|com.cn)$", r"\1", url.host
            ),
            "schema_name": url.database if url.database else "default",
        }
        opts.update(url.query)
        if "verify" in opts:
            verify = opts["verify"]
            try:
                verify = bool(strtobool(verify))
            except ValueError:
                # Probably a file name of the CA cert bundle to use
                pass
            opts.update({"verify": verify})

        return opts

    def do_ping(self, dbapi_connection):
        return dbapi_connection.test_connection()

    def get_schema_names(self, connection, **kw):
        # DynamoDB does not have the concept of a schema
        return ["default"]

    @reflection.cache
    def _get_tables(self, connection, schema=None, **kw):
        raw_connection = self._raw_connection(connection)
        with raw_connection.driver_connection.cursor() as cursor:
            return cursor.list_tables()

    def _get_column_type(self, metadata, attribute_name) -> str:
        col_type = types.NullType

        for attr in metadata["AttributeDefinitions"]:
            if attr["AttributeName"] == attribute_name:
                type_ = attr["AttributeType"]
                if type_ == "S":
                    col_type = types.String
                elif type_ == "N":
                    col_type = types.Numeric
                elif type_ == "B":
                    col_type = types.BINARY
                else:
                    util.warn(f"Did not recognize type '{type_}'")
                    col_type = types.NullType
        return col_type()

    def has_table(self, connection, table_name, schema=None, **kw):
        try:
            columns = self.get_columns(connection, table_name, schema)
            return True if columns else False
        except exc.NoSuchTableError:
            return False

    @reflection.cache
    def get_columns(self, connection, table_name, schema=None, **kw):
        raw_connection = self._raw_connection(connection)
        with raw_connection.driver_connection.cursor() as cursor:
            try:
                metadata_ = cursor.get_table_metadata(table_name)

                columns = [
                    {
                        "name": c["AttributeName"],
                        "type": self._get_column_type(metadata_, c["AttributeName"]),
                        "nullable": False,
                        "default": None,
                        "autoincrement": False,
                        "comment": None,
                        "dialect_options": {},
                    }
                    for c in metadata_["KeySchema"]
                ]

                return columns
            except OperationalError as e:
                cause = e.__cause__
                if (
                    isinstance(cause, botocore.exceptions.ClientError)
                    and cause.response["Error"]["Code"] == "ResourceNotFoundException"
                ):
                    raise exc.NoSuchTableError(table_name) from e
                raise

    def get_table_names(self, connection, schema=None, **kw):
        return self._get_tables(connection, schema, **kw)

    def get_view_names(self, connection, schema=None, **kw):
        return []

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        # DynamoDB has no support for foreign keys.
        return []  # pragma: no cover

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        # DynamoDB has no support for primary keys.
        return []  # pragma: no cover

    def get_indexes(self, connection, table_name, schema=None, **kw):
        # DynamoDB has no support for indexes.
        return []  # pragma: no cover

    def do_rollback(self, dbapi_connection):
        # No transactions for DynamoDB
        pass  # pragma: no cover

    def _check_unicode_returns(self, connection, additional_tests=None):
        # Requests gives back Unicode strings
        return True  # pragma: no cover

    def _check_unicode_description(self, connection):
        # Requests gives back Unicode strings
        return True  # pragma: no cover


class DynamoDBRestDialect(DynamoDBDialect):
    driver = "rest"
    supports_statement_cache = True
